// Based on https://codemirror.net/demo/search.html
// and https://codemirror.net/addon/search/search.js

// Search cursor is NEEDED
require('codemirror/addon/search/searchcursor');
// Show search matches on scrollbar (optional but I like it)
require('codemirror/addon/search/matchesonscrollbar');
require('codemirror/addon/search/matchesonscrollbar.css');

import CodeMirror = require('codemirror');

/**
 * Mostly providing a typed API on top of `search`
 */
export let commands = {
  search: (cm:CodeMirror.EditorFromTextArea, query: RegExp) => startSearch(cm, getSearchState(cm), query),
  hideSearch: (cm:CodeMirror.EditorFromTextArea) => hideSearch(cm),
  findNext: (cm:CodeMirror.EditorFromTextArea, query: RegExp) => findNextIfNotAlreadyDoing(cm,query,false),
  findPrevious: (cm:CodeMirror.EditorFromTextArea, query: RegExp) => findNextIfNotAlreadyDoing(cm,query,true),
  replaceNext: (cm: CodeMirror.EditorFromTextArea, newText: string) => simpleReplace(cm, newText, false),
  replacePrevious: (cm: CodeMirror.EditorFromTextArea, newText: string) => simpleReplacePrevious(cm, newText),
  replaceAll: (cm: CodeMirror.EditorFromTextArea, newText: string) => simpleReplace(cm, newText, true),
}

/** This is pulled out as is from `replace` function below. Just the dialog calls were removed 🌹 */
function simpleReplace(codeMirror: CodeMirror.EditorFromTextArea, newText: string, all: boolean) {

    /** These are the variables that I hand to introduce */
    let cm: any = codeMirror;
    let query = getSearchState(cm).query; // Note: this is different from what was there in that function (lastQuery)
    let text = newText;

    /**
     * The following lines are based on the replace function
    */

    if (all) {
        replaceAll(cm, query, text)
    } else {
        var cursor: CodeMirror.SearchCursor = getSearchCursor(cm, query, cm.getCursor("start"));

        /** Next match */
        var match;
        if (!(match = cursor.findNext())) {
            cursor = getSearchCursor(cm, query, CodeMirror.Pos(cm.firstLine(), 0));
            if (!(match = cursor.findNext()))
                return;
        }
        cm.setSelection(cursor.from(), cursor.to());
        cm.scrollIntoView({ from: cursor.from(), to: cursor.to() });
        doReplace();

        // Take them to the next match if any
        findNext(cm, false);
    }
    
    /** straight out of search function. Just trimed out as we know our query is a regex already */
    function doReplace() {
        cursor.replace(text.replace(/\$(\d)/g, function(_, i) { return match[i]; }));
    };
}

/** Based on simpleReplace defined above, but using `previous` search + no support for all */
function simpleReplacePrevious(codeMirror: CodeMirror.EditorFromTextArea, newText: string) {

    /** These are the variables that I hand to introduce */
    let cm: any = codeMirror;
    let query = getSearchState(cm).query; // Note: this is different from what was there in that function (lastQuery)
    let text = newText;

    /**
     * The following lines are based on the replace function
    */

    var cursor: CodeMirror.SearchCursor = getSearchCursor(cm, query, cm.getCursor("start"));

    /** Next match */
    var match;
    if (!(match = cursor.findNext())) {
        cursor = getSearchCursor(cm, query, CodeMirror.Pos(cm.firstLine(), 0));
        if (!(match = cursor.findPrevious()))
            return;
    }
    cm.setSelection(cursor.from(), cursor.to());
    cm.scrollIntoView({ from: cursor.from(), to: cursor.to() });
    doReplace();

    // Take them to the previous match if any
    findNext(cm, true);

    /** straight out of search function. Just trimed out as we know our query is a regex already */
    function doReplace() {
        cursor.replace(text.replace(/\$(\d)/g, function(_, i) { return match[i]; }));
    };
}

/** pulled from clearSearch below */
function hideSearch(cm) {cm.operation(function() {
  var state = getSearchState(cm);
  if (state.overlay){
      cm.removeOverlay(state.overlay);
  }
  if (state.annotate) { state.annotate.clear(); state.annotate = null; }
  return;
});}

function findNextIfNotAlreadyDoing(cm: CodeMirror.EditorFromTextArea, query: RegExp, rev = false) {
    var state = getSearchState(cm);
    if (!state.query || state.query.toString() !== query.toString()){
        startSearchNoOverlay(cm,getSearchState(cm), query);
    }
    findNext(cm,rev);
}

/** same as startSearch but with no overlay */
function startSearchNoOverlay(cm, state, query) {
  state.queryText = query;
  state.query = parseQuery(query);
  if (cm.showMatchesOnScrollbar) {
    if (state.annotate) { state.annotate.clear(); state.annotate = null; }
    state.annotate = cm.showMatchesOnScrollbar(state.query, queryCaseInsensitive(state.query));
  }
}

/**
 * See docs https://codemirror.net/doc/manual.html#addon_search
 * Mostly source code AS IT IS, but needed function local to the file (e.g. startSearch), so imported here
 * MODIFICATION BELOW
 * - parseQuery is commented out in favor of `parseQuery (x)=>x` as we already use a regex passed in
 *      Prevents the *magic detection* logic
 * - The commands need to be commented out othewise CM will intercept calls to `Ctrl+F`
 * - Add name:'searchOverlayMode' to the return of `searchOverlay` otherwise if errors happen CM will report "undefined mode failed to advance stream"
 * - Fix searchOverlay with https://github.com/codemirror/CodeMirror/pull/3619
 */

function searchOverlay(query, caseInsensitive) {
    if (typeof query == "string")
      query = new RegExp(query.replace(/[\-\[\]\/\{\}\(\)\*\+\?\.\\\^\$\|]/g, "\\$&"), caseInsensitive ? "gi" : "g");
    else if (!query.global)
      query = new RegExp(query.source, query.ignoreCase ? "gi" : "g");

    return {
      name:'searchOverlayMode',
      token: function(stream) {
      query.lastIndex = stream.pos;
      var match = query.exec(stream.string);
      if (match && match.index == stream.pos) {
        // https://github.com/codemirror/CodeMirror/pull/3619
        stream.pos += match[0].length || 1;
        return "searching";
      } else if (match) {
        stream.pos = match.index;
      } else {
        stream.skipToEnd();
      }
    }};
  }

  function SearchState() {
    this.posFrom = this.posTo = this.lastQuery = this.query = null;
    this.overlay = null;
  }

  function getSearchState(cm) {
    return cm.state.search || (cm.state.search = new SearchState());
  }

  function queryCaseInsensitive(query) {
    return typeof query == "string" && query == query.toLowerCase();
  }

  function getSearchCursor(cm, query, pos?) {
    // Heuristic: if the query string is all lowercase, do a case insensitive search.
    return cm.getSearchCursor(query, pos, queryCaseInsensitive(query));
  }

  function persistentDialog(cm, text, deflt, f) {
    cm.openDialog(text, f, {
      value: deflt,
      selectValueOnOpen: true,
      closeOnEnter: false,
      onClose: function() { clearSearch(cm); }
    });
  }

  function dialog(cm, text, shortText, deflt, f) {
    if (cm.openDialog) cm.openDialog(text, f, {value: deflt, selectValueOnOpen: true});
    else f(prompt(shortText, deflt));
  }

  function confirmDialog(cm, text, shortText, fs) {
    if (cm.openConfirm) cm.openConfirm(text, fs);
    else if (confirm(shortText)) fs[0]();
  }

  function parseString(string:string) {
    return string.replace(/\\(.)/g, function(_, ch) {
      if (ch == "n") return "\n"
      if (ch == "r") return "\r"
      return ch
    })
  }

  function parseQuery(query) {
    // MODIFICATION
    // bypassing any query parsing
    return query;

    // var isRE = query.match(/^\/(.*)\/([a-z]*)$/);
    // if (isRE) {
    //   try { query = new RegExp(isRE[1], isRE[2].indexOf("i") == -1 ? "" : "i"); }
    //   catch(e) {} // Not a regular expression after all, do a string search
    // } else {
    //   query = parseString(query)
    // }
    // if (typeof query == "string" ? query == "" : query.test(""))
    //   query = /x^/;
    // return query;
  }

  var queryDialog =
    'Search: <input type="text" style="width: 10em" class="CodeMirror-search-field"/> <span style="color: #888" class="CodeMirror-search-hint">(Use /re/ syntax for regexp search)</span>';

  function startSearch(cm, state, query) {
    state.queryText = query;
    state.query = parseQuery(query);
    cm.removeOverlay(state.overlay, queryCaseInsensitive(state.query));
    state.overlay = searchOverlay(state.query, queryCaseInsensitive(state.query));
    cm.addOverlay(state.overlay);
    if (cm.showMatchesOnScrollbar) {
      if (state.annotate) { state.annotate.clear(); state.annotate = null; }
      state.annotate = cm.showMatchesOnScrollbar(state.query, queryCaseInsensitive(state.query));
    }
  }

  function doSearch(cm, rev?, persistent?) {
    var state = getSearchState(cm);
    if (state.query) return findNext(cm, rev);
    var q = cm.getSelection() || state.lastQuery;
    if (persistent && cm.openDialog) {
      var hiding = null
      persistentDialog(cm, queryDialog, q, function(query, event) {
        CodeMirror.e_stop(event);
        if (!query) return;
        if (query != state.queryText) startSearch(cm, state, query);
        if (hiding) hiding.style.opacity = 1
        findNext(cm, event.shiftKey, function(_, to) {
          var dialog
          if (to.line < 3 && document.querySelector &&
              (dialog = cm.display.wrapper.querySelector(".CodeMirror-dialog")) &&
              dialog.getBoundingClientRect().bottom - 4 > cm.cursorCoords(to, "window").top)
            (hiding = dialog).style.opacity = .4
        })
      });
    } else {
      dialog(cm, queryDialog, "Search for:", q, function(query) {
        if (query && !state.query) cm.operation(function() {
          startSearch(cm, state, query);
          state.posFrom = state.posTo = cm.getCursor();
          findNext(cm, rev);
        });
      });
    }
  }

  function findNext(cm, rev, callback?) {cm.operation(function() {
    var state = getSearchState(cm);
    var cursor = getSearchCursor(cm, state.query, rev ? state.posFrom : state.posTo);
    if (!cursor.find(rev)) {
      cursor = getSearchCursor(cm, state.query, rev ? CodeMirror.Pos(cm.lastLine()) : CodeMirror.Pos(cm.firstLine(), 0));
      if (!cursor.find(rev)) return;
    }
    cm.setSelection(cursor.from(), cursor.to());
    cm.scrollIntoView({from: cursor.from(), to: cursor.to()}, 20);
    state.posFrom = cursor.from(); state.posTo = cursor.to();
    if (callback) callback(cursor.from(), cursor.to())
  });}

  function clearSearch(cm) {cm.operation(function() {
    var state = getSearchState(cm);
    state.lastQuery = state.query;
    if (!state.query) return;
    state.query = state.queryText = null;
    cm.removeOverlay(state.overlay);
    if (state.annotate) { state.annotate.clear(); state.annotate = null; }
  });}

  var replaceQueryDialog =
    ' <input type="text" style="width: 10em" class="CodeMirror-search-field"/> <span style="color: #888" class="CodeMirror-search-hint">(Use /re/ syntax for regexp search)</span>';
  var replacementQueryDialog = 'With: <input type="text" style="width: 10em" class="CodeMirror-search-field"/>';
  var doReplaceConfirm = "Replace? <button>Yes</button> <button>No</button> <button>All</button> <button>Stop</button>";

  function replaceAll(cm, query, text) {
    cm.operation(function() {
      for (var cursor = getSearchCursor(cm, query); cursor.findNext();) {
        if (typeof query != "string") {
          var match = cm.getRange(cursor.from(), cursor.to()).match(query);
          cursor.replace(text.replace(/\$(\d)/g, function(_, i) {return match[i];}));
        } else cursor.replace(text);
      }
    });
  }

  function replace(cm, all) {
    if (cm.getOption("readOnly")) return;
    var query = cm.getSelection() || getSearchState(cm).lastQuery;
    var dialogText = all ? "Replace all:" : "Replace:"
    dialog(cm, dialogText + replaceQueryDialog, dialogText, query, function(query) {
      if (!query) return;
      query = parseQuery(query);
      dialog(cm, replacementQueryDialog, "Replace with:", "", function(text) {
        text = parseString(text)
        if (all) {
          replaceAll(cm, query, text)
        } else {
          clearSearch(cm);
          var cursor = getSearchCursor(cm, query, cm.getCursor());
          var advance = function() {
            var start = cursor.from(), match;
            if (!(match = cursor.findNext())) {
              cursor = getSearchCursor(cm, query);
              if (!(match = cursor.findNext()) ||
                  (start && cursor.from().line == start.line && cursor.from().ch == start.ch)) return;
            }
            cm.setSelection(cursor.from(), cursor.to());
            cm.scrollIntoView({from: cursor.from(), to: cursor.to()});
            confirmDialog(cm, doReplaceConfirm, "Replace?",
                          [function() {doReplace(match);}, advance,
                           function() {replaceAll(cm, query, text)}]);
          };
          var doReplace = function(match) {
            cursor.replace(typeof query == "string" ? text :
                           text.replace(/\$(\d)/g, function(_, i) {return match[i];}));
            advance();
          };
          advance();
        }
      });
    });
  }

  // CodeMirror.commands.find = function(cm) {clearSearch(cm); doSearch(cm);};
  // CodeMirror.commands.findPersistent = function(cm) {clearSearch(cm); doSearch(cm, false, true);};
  // CodeMirror.commands.findNext = doSearch;
  // CodeMirror.commands.findPrev = function(cm) {doSearch(cm, true);};
  // CodeMirror.commands.clearSearch = clearSearch;
  // CodeMirror.commands.replace = replace;
  // CodeMirror.commands.replaceAll = function(cm) {replace(cm, true);};
